#!/usr/bin/env python
from gimpfu import *
import os
from os.path import expanduser
from os import listdir
import platform
import random
import time
import datetime
import colorsys

# preset
maxwh=3000
quota_to_auto_balance=50
quota_of_enhance=50  
quota_of_blacks_to_remove=24
quota_of_whites_to_remove=24


############################################################################
# Service routines
#---------------------------------------------------------------------------
############################################################################
# from Original drawable to Output image
#---------------------------------------------------------------------------
def createOutimage(inImage, inLayer,
                        widthheight,
                        ):
     
    owidth = inImage.width
    oheight = inImage.height
    otype = inImage.base_type
    name = inLayer.name
    if '.' in name:
        split = name.split(".", 1)
        name = split[0]
    inImage_name = inImage.name
    #gimp.message(str(inImage_name))
    split = inImage_name.split('(')   # when name comes from image
    inImage_name = split[0]
    #gimp.message(str(inImage_name))
    initial = inImage_name[0:1]
    if initial=="[":    # std name is [name]; name attributed by ofnuts' filter is name without [] 
        end = len(inImage_name)-2
        inImage_name = inImage_name[1:end]    # remove []
        #gimp.message(str(inImage_name))
    
   # define/calculate the scale factor for the (new) output image -------------------------
    realWidthHeight = owidth+oheight      # real size of the input image
    if widthheight == 0 :   # means the user wants to have output size same as input size
        if realWidthHeight<=maxwh:
            widthheight = realWidthHeight
        else:
            widthheight = maxwh  # max
            #gimp.message ("forced max width+height")
    optiWidthHeight = widthheight       # size to work on to get best results
    scale = (optiWidthHeight*1.0)/(realWidthHeight*1.0)      # ratio for scaling

    rwidth=int(owidth*scale)
    rheight=int(oheight*scale)
    
   # create a new image of the decided size ------------------------------------------------
   # --- copy the source layer
   # --- scale
    newImage = pdb.gimp_image_new (rwidth, rheight, otype)                    
    newBG = pdb.gimp_layer_new_from_drawable(inLayer, newImage)     # background layer
    newBG.mode = LAYER_MODE_NORMAL
    newBG.name = "NewImage: Background Layer from "+name
    newBG.opacity = 100.0 
    newImage.add_layer(newBG, 0)
    pdb.gimp_layer_scale(newBG, rwidth, rheight, False)
    pdb.gimp_layer_add_alpha (newBG)    # RGBA
    
   # keep a copy of the original background below the one which will be used ----------------
   # --- and set its visibility off (it should not enter in the process)
    Original = newBG.copy()
    newImage.add_layer(Original, 1)     # below
    pdb.gimp_item_set_visible (Original, False)
    Original.name = name+" (original)"
    
   # add to the BG some "equalization" of the colours values ------------
   # --- hence apply G'MIC auto-balance (replacing GIMP equalize),
   # --- set the opacity to quota_to_auto_balance,
   # --- merge down with the BG
    if int(quota_to_auto_balance)>0:
        pdb.plug_in_gmic_qt(newImage, newBG, 1,2 ,
        #[G'MIC] Auto Balance: gcd_auto_balance 30,0,0,1,0
        "-v - -gcd_auto_balance 30,0,0,1,0 ")
        newBGequ = newImage.layers[0]
        newBGequ.opacity=float(quota_to_auto_balance*1.0)    
        newBG=pdb.gimp_image_merge_down (newImage, newBGequ, CLIP_TO_IMAGE)
        newBG.name = "NewImage: BG with partially auto_balanced colours"
    
   # apply a simple Gimp color_enhance -----------------------------
   # --- hence duplicate again the BG, apply the Gimp plug-in-color-enhance,
   # --- set opacity to quota_of_enhanceand merge down
    if int(quota_of_enhance)>0:  
        newBGenh = newBG.copy()
        newImage.add_layer(newBGenh,0)
        newBGenh.opacity=float(quota_of_enhance)   # float reqd when called  
        newBGenh.mode = LAYER_MODE_NORMAL    
        pdb.plug_in_color_enhance (newImage, newBGenh)
        newBG = pdb.gimp_image_merge_down (newImage, newBGenh, CLIP_TO_BOTTOM_LAYER)
        newBG.name = "NewImage: BG after equalization with partially enhanced colours"
    
   # suppress almost black and almost white pixels ------------------
   # --- hence apply the Gimp levels such that the drawable
   # --- will have values starting from quota_of_blacks_to_remove to 255-quota_of_whites_to_remove
    level_start_value = float(quota_of_blacks_to_remove*1.0/255.0)
    level_end_value = 1.00-float(quota_of_whites_to_remove*1.0/255.0)
    #gimp.message ("min,max values="+str(level_start_value)+", "+str(level_end_value))
    if int(quota_of_blacks_to_remove)>0 or int(quota_of_whites_to_remove)>0:    
       #pdb.gimp_drawable_levels (newBG, HISTOGRAM_VALUE, 0.0, 1.0, False,
       #                              1.00, level_start_value, level_end_value, False)
       # better to use curves_spline !!!
        pdb.gimp_drawable_curves_spline (newBG, HISTOGRAM_VALUE, 4,
                                         (0.0,level_start_value,
                                          level_end_value, 1.0))   
        newBG.name = "NewImage: BG after equalization and color-enhance with blacks and pure white removed"
        
    newBG.opacity = 100.0
    
    return(newImage,Original,inImage_name)

# ADDITIONS taken from Ancillary: "DistressedPoster"
#---------------------------------------------------------------------------
# Poster thru Gmic: posterize using G'mic 
#---------------------------------------------------------------------------
def GmicPoster (inImage, inLayer, inImage_name,
                nr_of_colours,
                ):
   
   # check number of layers
    (nL,lL) = pdb.gimp_image_get_layers(inImage)
    
    width = inLayer.width
    height = inLayer.height
    WH = width+height
    
    # remove previously generated palettes
    Pname="Palette generated by "+inImage_name  # delete those related to this image
    PnameALL="Palette generated by "   # delete ALL generated by this function  10 June 2020
    (nP,lP)=pdb.gimp_palettes_get_list(PnameALL)
    if nP>0:
        x=0
        for x in range(0,nP): 
            pdb.gimp_palette_delete (lP[x])
        #gimp.message("deleted "+str(nP)+" palette(s) generated in previous run(s)")
        
    pdb.plug_in_gmic_qt(inImage, inLayer, 1,2 ,
    #[G'MIC] Posterize: fx_posterize 150,30,1,12,0,0,0,0,50,50
    "-v - -fx_posterize 150,30,1,"+str(nr_of_colours)+",0,0,0,0,50,50 ")
    PosterLayer = inImage.layers[0]
    PosterLayer.mode = LAYER_MODE_NORMAL
    PosterLayer.opacity = 100.0  
    PosterLayer.name = "[GMIC] -fx_posterize 150,30,1,"+str(nr_of_colours)+",0,0,0,0,50,50 "
    
    pdb.plug_in_gmic_qt(inImage, PosterLayer, 1,2 ,
    #[G'MIC] Colors to Layers: fx_split_colors 0,256,0,0
    "-v - -fx_split_colors 0,256,0,0 ")
    (num_layers,layer_ids) = pdb.gimp_image_get_layers (inImage)
    generated = num_layers-(nL+1)    # received ininput plus created PosterLayer
    #gimp.message ("generated "+str(generated)+" layers")
    
    gimp.context_push()
    areaNR=0
    area_color=[]   # empty list
    for areaNR in range(0,generated):
        #one_color=pdb.gimp_palette_entry_get_color(temp_palette, areaNR)
        pdb.gimp_context_set_antialias(False)
        pdb.gimp_context_set_feather(False)
        pdb.gimp_context_set_sample_criterion(SELECT_CRITERION_COMPOSITE)
        pdb.gimp_context_set_sample_merged(False)
        pdb.gimp_context_set_sample_threshold(0.0)
        pdb.gimp_context_set_sample_transparent(True)
        # select transparent
        pdb.gimp_image_select_color(inImage, CHANNEL_OP_REPLACE,
                                    inImage.layers[areaNR], (0,0,0,0)) 
        pdb.gimp_selection_invert (inImage)
        (there_is_a_selection, x1, y1, x2, y2) = pdb.gimp_selection_bounds(inImage)
        if there_is_a_selection==True:
            (meanR,B,C,D,E,F) = pdb.gimp_drawable_histogram (inImage.layers[areaNR],
                                                         HISTOGRAM_RED, 0.0, 1.0)
            (meanG,B,C,D,E,F) = pdb.gimp_drawable_histogram (inImage.layers[areaNR],
                                                         HISTOGRAM_GREEN, 0.0, 1.0)
            (meanB,B,C,D,E,F) = pdb.gimp_drawable_histogram (inImage.layers[areaNR],
                                                         HISTOGRAM_BLUE, 0.0, 1.0)
            color = (int(meanR),int(meanG),int(meanB))  # WARNING: int() is necessary !
            area_color.append(color)
            #gimp.message ("color for area "+str(areaNR)+" is "+str(color))
        else:
            gimp.message ("area nr. "+str(areaNR)+" there_is_NO_selection")
            
    pdb.gimp_selection_none(inImage)
     
    generated_palette = pdb.gimp_palette_new (Pname)
    for areaNR in range(0,generated):
        pdb.gimp_palette_add_entry(generated_palette, str(areaNR), area_color[areaNR])
    #gimp.message ("created: "+str(generated_palette))
    
    pdb.gimp_palettes_refresh()
    gimp.context_pop()
    #for areaNR in range(0,generated):
    #    pdb.gimp_image_remove_layer (inImage,inImage.layers[areaNR])
                                     
    return (PosterLayer, generated_palette)
    
# END of ADDITIONS
############################################################################
# texturize the image
#---------------------------------------------------------------------------
def texturizeImage(inImage, inLayer, doDistress,
                     temp_palette,
                     patternsetsingle):
   
   # check number of layers
    (nL,lL) = pdb.gimp_image_get_layers(inImage)
    if nL<2:
        gimp.message ("this block should receive in input an image with at least 2 layers, quantized")
        return
    
    # get OS and generate appropriate slash
    home = expanduser("~")
    # additions for compatibility with Linux systems
    myplatform=platform.system()
    #pdb.gimp_message("I'm running "+myplatform)
    if myplatform=="Linux":
        myslash="/"
    if myplatform=="Windows":
        myslash="\\"
    if myplatform=="Darwin":    # Apple (addition 03/12*2017)
        myslash="/"
    
    # environmental set-ups
    pdb.gimp_context_set_sample_criterion (SELECT_CRITERION_COMPOSITE)
    pdb.gimp_context_set_sample_threshold (0.01)                       # important !!!
    pdb.gimp_context_set_sample_transparent (False)
    pdb.gimp_context_set_sample_merged(False)
    # constants for the rotations
    ROTATE_FORW1 = ROTATE_90; ROTATE_BACK1 = ROTATE_270   # fixed
    ROTATE_FORW2 = ROTATE_180; ROTATE_BACK2 = ROTATE_180  # fixed
    ROTATE_FORW3 = ROTATE_270; ROTATE_BACK3 = ROTATE_90   # fixed
    
   # check to have the patterns group active in your patterns sets -------------------
    patternsetchosen=patternsetsingle                # single set 
    LP=[]
    (nP,LP) = pdb.gimp_patterns_get_list(patternsetchosen)
    #gimp.message ("patterns found = "+str(nP)+" using patternsetchosen "+str(patternsetchosen))
    if nP == 0:
        gimp.message ("change discriminant for patterns name, no pattern found; exit")
        return
    
    # image dimension
    widthheight=inImage.width+inImage.height
    
    # output image
    wrkImage=inImage.duplicate()
    
    nr_areas = pdb.gimp_palette_get_info(temp_palette)
    quantxtext_layer = wrkImage.layers[0]
    quantxtext_layer.name = "WrkImage: Quantized Layer"
    quantized_BG = wrkImage.layers[1]
    quantized_BG.name = "WrkImage: Quantized BG"
     
   #  Enter a loop to fill the N areas with appropriate patterns ------------------                      

    pdb.gimp_image_undo_freeze(wrkImage)                    # save time and space
    
    pdb.gimp_selection_all (wrkImage)
    (a,b,c,LayerPixels,e,f) = pdb.gimp_drawable_histogram (quantized_BG,
                              HISTOGRAM_VALUE, 0.0, 1.0)
    #gimp.message ("LayerPixels="+str(LayerPixels))
    
    areaNR=0
    
   # --- for each colour area: 
    for areaNR in range(0,nr_areas):   #(0,nr_areas):
        one_color=pdb.gimp_palette_entry_get_color(temp_palette, areaNR)
        # rotate moved up here
        if areaNR%4<>0:
            if areaNR%4==1:
                pdb.gimp_image_rotate (wrkImage, ROTATE_FORW1)
            elif areaNR%4==2:
                pdb.gimp_image_rotate (wrkImage, ROTATE_FORW2)
            elif areaNR%4==3:
                pdb.gimp_image_rotate (wrkImage, ROTATE_FORW3)                 
        pdb.gimp_image_select_color(wrkImage, CHANNEL_OP_REPLACE, quantized_BG, one_color)
        (AreaMeanL,b,c,d,e,f) = pdb.gimp_drawable_histogram (quantized_BG,
                                          HISTOGRAM_LUMINANCE, 0.0, 1.0)
        (a,b,c,AreaPixels,e,f) = pdb.gimp_drawable_histogram (quantized_BG,
                                          HISTOGRAM_VALUE, 0.0, 1.0)
        #gimp.message ("AreaPixels="+str(AreaPixels))
        areaquota = float(AreaPixels*1.0/LayerPixels*1.0)   # on VALUE
        #gimp.message ("for color "+str(one_color)+" areaquota="+str(areaquota))
   # --- identify the hue of the area 
        (R,G,B,A)=one_color
        #gimp.message ("Area colour = "+str(one_color))
   # --- convert RGB to HSV
   # --- get hsv percentage: range (0-1, 0-1, 0-1)
        color_hsv_percentage=colorsys.rgb_to_hsv(R/255.0, G/255.0, B/255.0) 
        #gimp.message (str(color_hsv_percentage))
        
   # --- choose a random pattern from the set 
        indexNR = random.randint(0,nP-1)
        pdb.gimp_context_set_pattern(LP[indexNR])  # set the pattern
        #gimp.message ("chosen pattern = "+str(LP[indexNR]))
   # --- open the pattern as image
        (Pwidth,Pheight,Pbpp) = pdb.gimp_pattern_get_info (LP[indexNR])
        Pimage = pdb.gimp_image_new (Pwidth, Pheight, RGB)
        Player = pdb.gimp_layer_new (Pimage, Pwidth, Pheight, RGBA_IMAGE, LP[indexNR],
                                     100.0, LAYER_MODE_NORMAL)
        Pimage.add_layer (Player,0)
        pdb.gimp_edit_fill (Player, FILL_PATTERN)
        (there_is_a_selection, x1, y1, x2, y2) = pdb.gimp_selection_bounds(wrkImage)
        areawidth=x2-x1; areaheight=y2-y1; areaWH=areawidth+areaheight
        (PatMeanL, b,c,d,e,f) = pdb.gimp_drawable_histogram (Player, HISTOGRAM_LUMINANCE, 0.0, 1.0)
        #gimp.message ("pattern vs area luma: "+str(PatMeanL)+" vs "+str(AreaMeanL))
        
   # --- rescale it to the desired size,
   # --- (which could be fixed -400 is good- or depending on the area amplitude -from 300 to 800-)
   # --- save the rescaled pattern image as a temporary pattern ("temp_pattern", for instance)
       
        # auto adjustable pattern size
        (a,b,c,d,count,f) = pdb.gimp_drawable_histogram (quantized_BG, HISTOGRAM_VALUE, 0.0, 1.0)
        SQRT = int(math.sqrt(count))   # proportional
        #gimp.message ("SQRT="+str(SQRT))
        ratioW=areawidth*1.0/areaWH; ratioH=areaheight*1.0/areaWH
        #gimp.message ("ratioW,ratioH="+str(ratioW)+","+str(ratioH))
        AW=int(float(ratioW*SQRT*1.0))
        AH=int(float(ratioH*SQRT*1.0))
        #gimp.message ("AW*AH="+str(AW)+"*"+str(AH))
        # don't scale down big patterns
        if Pwidth<AW:
            pdb.gimp_image_scale (Pimage, AW, AH)
            #gimp.message ("orig W="+str(Pwidth)+", scaled W="+str(AW))  # check
        # adjust pattern luma to area lume
        meanLratio = PatMeanL*1.0/AreaMeanL*1.0
        if meanLratio > 0.9 and meanLratio < 1.1: # it's OK
            adj_levels = 0 # leave as it is
            #gimp.message ("mean before="+str(mean)+", no need to apply levels")
        elif meanLratio < 0.9:  # it's dark
            adj_levels = 1 # brighten needed
            run=0
            while meanLratio < 0.9 and run<9:
                pdb.gimp_drawable_levels (Player, HISTOGRAM_VALUE, 0.0, 1.0, False,
                                          1.11, 0.0, 1.0, False)   # apply a moderate brighten
                (PatMeanL, b,c,d,e,f) = pdb.gimp_drawable_histogram (Player, HISTOGRAM_LUMINANCE,
                                                                     0.0, 1.0)
                meanLratio = PatMeanL*1.0/AreaMeanL*1.0
                run+=1    
            #gimp.message ("pattern mean after applying Levels because it was too dark="+str(PatMeanL))
        elif meanLratio > 1.11:  # it's light
            adj_levels = 2 # darken
            run=0
            while meanLratio > 1.11 and run<9:
                pdb.gimp_drawable_levels (Player, HISTOGRAM_VALUE, 0.0, 1.0, False,
                                          0.90, 0.0, 1.0, False)   # apply a moderate darken
                (PatMeanL, b,c,d,e,f) = pdb.gimp_drawable_histogram (Player, HISTOGRAM_LUMINANCE,
                                                                     0.0, 1.0)
                meanLratio = PatMeanL*1.0/AreaMeanL*1.0
                run+=1
            #gimp.message ("pattern mean after applying Levels because it was too light="+str(PatMeanL))
        # colorize to area color
        pdb.gimp_drawable_colorize_hsl (Player, color_hsv_percentage[0],color_hsv_percentage[1],0.0)
        # end luma of pattern adjustment
        d_dir = pdb.gimp_gimprc_query ("gimp_dir")
        Tpat_name = "temp_pat"
        Tpat_name2 = d_dir+myslash+"patterns"+myslash+Tpat_name+".pat"        # changed for compatibility with Linux OS
        pdb.file_pat_save (Pimage, Player, Tpat_name2, Tpat_name, "temp_pat")
        pdb.gimp_patterns_refresh()
        (there_is_a_selection, x1, y1, x2, y2) = pdb.gimp_selection_bounds(wrkImage)
        
   # --- fill the colour area with that (rescaled) pattern        
        if there_is_a_selection==True:
            if doDistress>0:
                if doDistress==1: #soft
                    inSpread=3.0; inGranularity=2.0; inSmooth=2.0
                if doDistress==2: #hard
                    inSpread=6.0; inGranularity=4.0; inSmooth=2.0
                if areaquota>float(0.25/nr_areas):
                    pdb.script_fu_distress_selection (wrkImage, quantxtext_layer,
                              127.0, inSpread, inGranularity, inSmooth, True, True)
                #else: gimp.message ("no distress applied to a small area")    
            pdb.gimp_context_set_pattern (Tpat_name)    
            pdb.gimp_edit_fill (quantxtext_layer, FILL_PATTERN)
        else:
            gimp.message_text=("area nr. "+str(areaNR)+" color "+str(one_color)+" there_is_NO_selection")
             
        ### new: adjust luminosity of the area
         
        divisor=8.0
        if (R+G+B) < ((256/divisor)*2.5):
            pdb.gimp_drawable_curves_spline (quantxtext_layer,HISTOGRAM_VALUE,
                                             6,(0.0,0.0, 0.5,0.4, 1.0,1.0))
        if (R+G+B) > ((256/divisor)*3.5):
            pdb.gimp_drawable_curves_spline (quantxtext_layer,HISTOGRAM_VALUE,
                                             6,(0.0,0.0, 0.5,0.6, 1.0,1.0))  
        ### end adjust luminosity
        #gimp.message ("areaNR%4="+str(areaNR%4))  
        if areaNR%4<>0:
            if areaNR%4==1:
                pdb.gimp_image_rotate (wrkImage, ROTATE_BACK1)
            elif areaNR%4==2:
                pdb.gimp_image_rotate (wrkImage, ROTATE_BACK2)
            elif areaNR%4==3:
                pdb.gimp_image_rotate (wrkImage, ROTATE_BACK3)                 
        #pdb.gimp_display_new (wrkImage)
   # end of the loop, set this layer mode to LINEAR_LIGHT
    #gimp.message ("end of areas loop")
    
    quantxtext_layer.name = "WrkImage: Quantized Layer with texture patterns of group "+patternsetchosen
    quantxtext_layer.mode = LAYER_MODE_GRAIN_MERGE  #LUMINANCE #HSV_VALUE #MULTIPLY #LINEAR_LIGHT
    pdb.gimp_selection_none(wrkImage)
    
   # at the end of the loop, check the luminosity of the fully "mosaicized" layer
   # and apply some adjustment using Gimp levels, if necessary 
    (mean, b,c,d,e,f) = pdb.gimp_drawable_histogram (quantxtext_layer, HISTOGRAM_LUMINANCE, 0.0, 1.0)
    #gimp.message ("mean before="+str(mean))
    if mean > 110 and mean < 161: # it's OK
        adj_levels = 0 # leave as it is
        #gimp.message ("mean before="+str(mean)+", no need to apply levels")
    elif mean < 111:  # it's dark
        adj_levels = 1 # brighten
        run=0
        while mean < 111 and run<9:
            pdb.gimp_drawable_levels (quantxtext_layer, HISTOGRAM_VALUE, 0.0, 1.0, False,
                                      1.25, 0.0, 1.0, False)   # apply a moderate brighten
            (mean, b,c,d,e,f) = pdb.gimp_drawable_histogram (quantxtext_layer, HISTOGRAM_LUMINANCE, 0.0, 1.0)
            run+=1    
        #gimp.message ("mean after applying Levels because it was too dark="+str(mean))
    elif mean > 160:  # it's light
        adj_levels = 2 # darken
        run=0
        while mean > 160 and run<9:
            pdb.gimp_drawable_levels (quantxtext_layer, HISTOGRAM_VALUE, 0.0, 1.0, False,
                                      0.80, 0.0, 1.0, False)   # apply a moderate darken
            (mean, b,c,d,e,f) = pdb.gimp_drawable_histogram (quantxtext_layer, HISTOGRAM_LUMINANCE, 0.0, 1.0)
            run+=1
        #gimp.message ("mean after applying Levels because it was too light="+str(mean))
    if adj_levels<>0:    
        quantxtext_layer.name = "WrkImage: Quantized & Texturized Layer after Gimp Levels adjustment "
    
   # often is good to remove also the "black-ish" and the whitish pixels
   # (using Gimp levels, for instance setting the output range from 0.1 to 0.9)    
    pdb.gimp_drawable_levels (quantxtext_layer, HISTOGRAM_VALUE, 0.0, 1.0, False,
                                      1.00, 0.15, 0.85, False)
    quantxtext_layer.name = "WrkImage: Quantized & Texturized Layer "
    
    #pdb.gimp_display_new (wrkImage) # for test
    
    textLayer = pdb.gimp_layer_new_from_drawable (quantxtext_layer, inImage)
    inImage.add_layer (textLayer, 0)
    textLayer.name = "NewImage: Texturized Layer (from WrkImage)"
    pdb.gimp_image_remove_layer (inImage, inLayer)
   # --- at this point the WORK image is no more needed
    pdb.gimp_image_delete (wrkImage)  # no need anymore
    
    return(inImage,temp_palette, patternsetchosen) 

############################################################################
# Finalize a quantized image
#---------------------------------------------------------------------------
def finalizeImage(inImage, inLayer,
                  doFlatten, restoreGrain,
                  showCanvas
                 ):
   
   # check number of layers
    (nL,lL) = pdb.gimp_image_get_layers(inImage)
    if nL<6:
        gimp.message ("this block should receive in input an image with at least 6 layers")
        return
    #gimp.message ("this block has correctly received in input an image with 6+ layers")
    texturized_layer = inImage.layers[0]
    new_BG = inImage.layers[4]
    original = inImage.layers[5]
    rwidth = inImage.width
    rheight = inImage.height
   # get back colours from the image BG:
   # --- copy the BG to create a layer ("for mising Hue/Colour")
    # add BG for original colours
    forHue = new_BG.copy()    
    inImage.add_layer (forHue, 0)
    forHue.name = "copy of BG for mixing Hue/Colour"
    forHue.mode = LAYER_MODE_BURN   #HSL_COLOR  #LCH_COLOR  
    forHue.opacity = 30.0   #60.0  #50.0   
    #gimp.message ("BG on top for hue adjust applied") 
   # create a layer from visible and set its opacity to 33.0    
    visible = pdb.gimp_layer_new_from_visible (inImage, inImage, "Visible")
    inImage.add_layer(visible, 0)
    visible.opacity = 33.0
    Outcome = pdb.gimp_layer_new_from_visible (inImage, inImage, "First Outcome")
    inImage.add_layer(Outcome,0)
    #  
    (finalmean,b,c,d,e,f) = pdb.gimp_drawable_histogram(Outcome,HISTOGRAM_LUMINANCE,0.0,1.0)
    #gimp.message ("final mean calculated "+str(finalmean)) 
    (originalmean,b,c,d,e,f) = pdb.gimp_drawable_histogram(original,HISTOGRAM_LUMINANCE,0.0,1.0)
    #gimp.message ("original mean calculated "+str(originalmean))  
    if finalmean > (originalmean*1.11):
        #gimp.message ("outcome layer darkening curves to be applied") 
        pdb.gimp_drawable_curves_spline(Outcome,HISTOGRAM_VALUE,6,(0.0,0.0, 0.5,0.45, 1.0,1.0))
        # HISTOGRAM_LUMINANCE abends !!!
    elif finalmean < (originalmean*0.90):
        #gimp.message ("outcome layer brightening curves applied") 
        pdb.gimp_drawable_curves_spline(Outcome,HISTOGRAM_VALUE,6,(0.0,0.0, 0.45,0.5, 1.0,1.0))
        # HISTOGRAM_LUMINANCE abends !!!
        
    #gimp.message ("FIRST OUTCOME (RAW) completed")
    if showCanvas > 0:
        Base2 = pdb.gimp_layer_new_from_visible (inImage, inImage, "")
        inImage.add_layer (Base2, 0)
        pdb.gimp_context_set_background ((129,128,125,255))
        pdb.gimp_drawable_fill (Base2, FILL_BACKGROUND)
        pdb.plug_in_gmic_qt(inImage, Base2, 1,0,       
        "-v - -texturize_canvas 20,4,0.8 ")
        Canvas = inImage.layers[0]
        Canvas.mode = LAYER_MODE_LINEAR_LIGHT     
        Canvas.opacity = 95.0    
        Canvas.name = "canvas 20,4,0.8 "
        if showCanvas==2: # hard
            Radius=3.0;Amount=2.5; Threshold=0.0       # strong sharpen canvas texture 
            pdb.plug_in_unsharp_mask (inImage, Canvas, Radius, Amount, Threshold)
            Canvas.name = "texturize_canvas 20,4,0.8 + USM 3,2.5 "
     
    #gimp.message ("show canvas processed")
    #  -------------------------------------------
    
    FinalOutcome = pdb.gimp_layer_new_from_visible (inImage, inImage, "Final Visible")
    inImage.add_layer(FinalOutcome, 0)
     
    #if doFlatten==True:
    #    FinalOutcome = pdb.gimp_image_merge_visible_layers (inImage, CLIP_TO_IMAGE)
        
    if int(quota_of_blacks_to_remove)>0:
        level_start_value = float(quota_of_blacks_to_remove*1.0/255.0)
    if int(quota_of_whites_to_remove)>0:
        level_end_value = 1.00-float(quota_of_whites_to_remove*1.0/255.0)
    if int(quota_of_blacks_to_remove)>0 or int(quota_of_whites_to_remove)>0:    
        pdb.gimp_drawable_levels (FinalOutcome, HISTOGRAM_VALUE, 0.0, 1.0, False,
                                      1.00, level_start_value, level_end_value, False)  
    #gimp.message ("quotas B&W processed")
    # post-processing: restore xx% of original "grain"
    originalOnTop = original.copy() #new_BG.copy()
    inImage.add_layer(originalOnTop, 0)
    pdb.gimp_item_set_visible (originalOnTop, True)
    originalOnTop.mode = LAYER_MODE_HARDLIGHT   #SCREEN  #MULTIPLY    #GRAIN_MERGE
    #pdb.gimp_drawable_hue_saturation (originalOnTop, HUE_RANGE_ALL,
    #                                  0.0, 0.0, -70.0, 0.0) # added 24may
    originalOnTop.name = "Original on Top for Grain"
    #gimp.message ("restore grain % = "+str(restoreGrain))
    originalOnTop.opacity = (restoreGrain*1.0)    # from 0.0 to 100.0
    #FinalOutcome = pdb.gimp_image_merge_down (inImage, originalOnTop, CLIP_TO_IMAGE)    
     
    return (inImage)

############################################################################
# Render a drawable a free strokes painting  
#---------------------------------------------------------------------------
def FreePainting2022A(inImage, inLayer,
                     widthheight, 
                     nr_of_colours, doDistress,
                     patternsetsingle, sharptexture, 
                     doFlatten, restoreGrain,
                     showCanvas):
    
    name = inLayer.name
    if '.' in name:
        split = name.split(".", 1)
        name = split[0]
    
    ts = time.time()
    st = datetime.datetime.fromtimestamp(ts).strftime('%Y-%m-%d %H-%M-%S')
    
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2) # error console
    
    # check to have the patterns group active in your patterns sets -------------------
    patternsetchosen=patternsetsingle                # single set 
    LP=[]
    (nP,LP) = pdb.gimp_patterns_get_list(patternsetchosen)
    #gimp.message ("patterns found = "+str(nP)+" using patternsetchosen "+str(patternsetchosen))
    if nP == 0:
        gimp.message ("change discriminant for patterns name, no pattern found; exit")
        return
    
    (newImage,original,newImage_name) = createOutimage (inImage, inLayer, 
                                        widthheight) 
    newLayer = newImage.layers[0]
    #gimp.message (newImage_name)
    
    #pdb.gimp_display_new (newImage)
    #return
    
    # Apply a Paint look
    # thru Segmentation
    if doDistress==0:
        edge=4.0; smooth=1.0;
    if doDistress==1: #soft
        edge=4.0; smooth=2.0;
    if doDistress==2: #hard
        edge=4.0; smooth=4.0;
    pdb.plug_in_gmic_qt(newImage, newLayer, 1,0,
    '-v - -fx_segment_watershed '+str(edge)+','+str(smooth)+'0,0,0,50,50 ')                    
   #'-v - -fx_segment_watershed 4,4,0,0,0,50,50 ')  # was
    newLayer = newImage.layers[0]
    newLayer.mode = LAYER_MODE_NORMAL
    newLayer.opacity = 100.0
    newLayer.name = "New BG Layer painted thru Segmentation"
    #pdb.plug_in_unsharp_mask (newImage, Painted, 3.0, 3.0, 0)
    
    #pdb.gimp_display_new (newImage)
    #return
     
    # Work images for GmicPoster <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
    wrkImageQ = pdb.gimp_image_duplicate (newImage)
    wrkLayerQ = wrkImageQ.layers[0]
    wrkImageH = pdb.gimp_image_duplicate (newImage)
    wrkLayerH = wrkImageH.layers[0]
    wrkImageF = pdb.gimp_image_duplicate (newImage)
    wrkLayerF = wrkImageH.layers[0]
     
    # POSTER 1/4 use GmicPoster <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
    (wrkPosterQ,createdpaletteQ) = GmicPoster(wrkImageQ, wrkLayerQ, newImage_name,
                                              int(nr_of_colours/4),
                                              )
     
    quantLayerQ = newLayer.copy()
    newImage.add_layer (quantLayerQ, 0)
    pdb.gimp_edit_copy (wrkPosterQ)
    floatsel = pdb.gimp_edit_paste (quantLayerQ, True)
    pdb.gimp_floating_sel_anchor (floatsel)
    quantxtextQ = quantLayerQ.copy()# INPUT for next step TEXTURIZE
    newImage.add_layer (quantxtextQ, 0)
    quantxtextQ.name = "input for texturize Q"
    
    #pdb.gimp_display_new (newImage)
    #return
    
    #gimp.message ("before texturize")
    (newImage,usedpalette,patternsetchosen) = texturizeImage(newImage, quantxtextQ,
                                              doDistress, createdpaletteQ,
                                              patternsetsingle)
    
    textLayerQ = newImage.layers[0]
    
    OutcomeQ = pdb.gimp_image_merge_down (newImage, textLayerQ, CLIP_TO_IMAGE)
    OutcomeQ.name = "Outcome Q"
     
    #pdb.gimp_display_new (newImage)
    #return
    
    # POSTER 1/2 use GmicPoster <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
    (wrkPosterH,createdpaletteH) = GmicPoster(wrkImageH, wrkLayerH, newImage_name,
                                              int(nr_of_colours/2),
                                              )
     
    #pdb.gimp_display_new (wrkImageH)
    
    quantLayerH = newLayer.copy()
    newImage.add_layer (quantLayerH, 0)
    pdb.gimp_edit_copy (wrkPosterH)
    floatsel = pdb.gimp_edit_paste (quantLayerH, True)
    pdb.gimp_floating_sel_anchor (floatsel)
    quantxtextH = quantLayerH.copy()# INPUT for next step TEXTURIZE
    newImage.add_layer (quantxtextH, 0)
    quantxtextH.name = "input for texturize H"
    
    #pdb.gimp_display_new (newImage)
    #return
    
    #gimp.message ("before texturize")
    (newImage,usedpalette,patternsetchosen) = texturizeImage(newImage, quantxtextH,
                                              doDistress, createdpaletteH,
                                              patternsetsingle)
    
    textLayerH = newImage.layers[0]
    
    OutcomeH = pdb.gimp_image_merge_down (newImage, textLayerH, CLIP_TO_IMAGE)
    OutcomeH.name = "Outcome H"
     
    #pdb.gimp_display_new (newImage)
    #return
     
    # POSTER full use GmicPoster <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
    (wrkPosterF,createdpaletteF) = GmicPoster(wrkImageF, wrkLayerF, newImage_name,
                                              nr_of_colours,
                                              )
     
    #pdb.gimp_display_new (wrkImageF)
    
    quantLayerF = newLayer.copy()
    newImage.add_layer (quantLayerF, 0)
    pdb.gimp_edit_copy (wrkPosterF)
    floatsel = pdb.gimp_edit_paste (quantLayerF, True)
    pdb.gimp_floating_sel_anchor (floatsel)
    quantxtextF = quantLayerF.copy()# INPUT for next step TEXTURIZE
    newImage.add_layer (quantxtextF, 0)
    quantxtextF.name = "input for texturize F"
    
    #pdb.gimp_display_new (newImage)
    #return
    
    #gimp.message ("before texturize")
    (newImage,usedpalette,patternsetchosen) = texturizeImage(newImage, quantxtextF,
                                              doDistress, createdpaletteF,
                                              patternsetsingle)
    
    textLayerF = newImage.layers[0]
    
    OutcomeF = pdb.gimp_image_merge_down (newImage, textLayerF, CLIP_TO_IMAGE)
    OutcomeF.name = "Outcome F"
     
    #pdb.gimp_display_new (newImage)
    #return
    
    if sharptexture==1: #soft
        pdb.plug_in_unsharp_mask (newImage, OutcomeQ, 3.0, 1.5, 0)
        pdb.plug_in_unsharp_mask (newImage, OutcomeH, 3.0, 1.5, 0)
        pdb.plug_in_unsharp_mask (newImage, OutcomeF, 3.0, 1.5, 0)
    elif sharptexture==2: #hard
        pdb.plug_in_unsharp_mask (newImage, OutcomeQ, 3.0, 3.0, 0)
        pdb.plug_in_unsharp_mask (newImage, OutcomeH, 3.0, 3.0, 0)
        pdb.plug_in_unsharp_mask (newImage, OutcomeF, 3.0, 3.0, 0)
    OutcomeQ.opacity = 25.0
    OutcomeH.opacity = 33.0
    OutcomeF.opacity = 50.0    
    OutcomeQ.mode = LAYER_MODE_MULTIPLY #GRAIN_MERGE
    OutcomeH.mode = LAYER_MODE_MULTIPLY #GRAIN_MERGE
    OutcomeF.mode = LAYER_MODE_MULTIPLY #GRAIN_MERGE
    textLayer = pdb.gimp_layer_new_from_visible (newImage, newImage,
                                                 "Merged texturized layer")
    newImage.add_layer (textLayer, 0)
    if sharptexture==1: #soft
        pdb.plug_in_unsharp_mask (newImage, textLayer, 3.0, 1.5, 0)
    elif sharptexture==2: #hard
        pdb.plug_in_unsharp_mask (newImage, textLayer, 3.0, 3.0, 0)
     
    #pdb.gimp_display_new (newImage)
    #return
    
    #gimp.message ("before finalize")
    finalizeImage(newImage, textLayer, 
                  doFlatten, restoreGrain,
                  showCanvas)
     
    #pdb.gimp_display_new (newImage)
    #return
    
    if doFlatten==True:
        finalLayer = pdb.gimp_image_merge_visible_layers (newImage,
                                                          CLIP_TO_IMAGE)
    else:
        finalLayer = pdb.gimp_layer_new_from_visible (newImage,
                                                      newImage, "")
        newImage.add_layer (finalLayer, 0)
    finalLayer.name=st+" "+name+", as a free paint (2022A)"
    finalLayer.name=finalLayer.name+", patterns set="+patternsetchosen
    finalLayer.name=finalLayer.name+", "+str(int(nr_of_colours))+" colours"
    textlook = ["basic","soft","hard"]
    textused = ", texture "+textlook[sharptexture]
    finalLayer.name=finalLayer.name+textused 
    if doFlatten==False:
        pdb.gimp_item_set_visible (finalLayer, False)   # to see effect of changes to ther layers
    
    pdb.gimp_display_new (newImage)
    
    pdb.gimp_message_set_handler(currenth) # current
    
    return (newImage)
     

# This is the plugin registration function  
register(
    "FreePainting2022A",
    "Apply free painting strokes (2022) (A)",
    "Free painting.",
    "Diego", 
    "Diego Nassetti ", 
    "2020/2022",
    "Free painting 2022 v.A",
    "RGB*",
    [
        (PF_IMAGE, "image", "Input image", None),
        (PF_DRAWABLE, "drawable", "Input drawable", None),
        (PF_ADJUSTMENT, "newsize", "Width+Height of New Image"
                                    +"\n (0=same as input, 6000=max)", 3000, (0, 6000, 100)),
        (PF_ADJUSTMENT, "colours", "How many colours max (8-64)", 15, (8, 64, 4)),
        (PF_OPTION, "distress", "Colour area distress", 1, ["basic", "soft", "hard"]),
        (PF_STRING, "patterns", "Single pattern set", "FPSgray1"),
        (PF_OPTION, "sharpen", "Textures visibility", 1, ["basic", "soft", "hard"]),
        (PF_TOGGLE, "flatten", "Flatten Image?", False),
        (PF_ADJUSTMENT, "restore", "% original grain over final Image?", 35.0,
                                                        (0.0, 100.0, 1.0)),
        (PF_OPTION, "canvas", "Canvas look", 1, ["*none", "soft", "hard"]),

        ],
    [
        (PF_IMAGE, "outimage", "Output image", None),
        ],
    FreePainting2022A,
    menu="<Image>/Diego/Artistic",
    )

main()
